#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 14 12:38:27 2021

@author: lorenzo
"""
import matplotlib 
import matplotlib.pyplot as plt

def update_fontsize(size):
    matplotlib.rcParams.update({'font.size': size})
    return
